@@ -29,6 +29,7 @@
 #include "htmlgdkpainter.h"
 #include "htmlprinter.h"
 #include "htmliframe.h"
+#include "htmlengine.h"
 #include "htmlengine-search.h"
 #include "htmlengine-save.h"
 #include "htmlsearch.h"
@@ -40,7 +41,6 @@
 
 HTMLIFrameClass html_iframe_class;
 static HTMLEmbeddedClass *parent_class = NULL;
-static gboolean calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed_objs);
 	
 static void
 iframe_set_base (GtkHTML *html, const char *url, gpointer data)
@@ -141,8 +141,10 @@ set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
 {
 	HTMLEngine *e = GTK_HTML (HTML_IFRAME (o)->html)->engine;
 
-	o->max_width = max_width;
-	html_object_set_max_width (e->clue, e->painter, max_width - (html_engine_get_left_border (e) + html_engine_get_right_border (e)));
+	if (o->max_width != max_width) {
+		o->max_width = max_width;
+		html_object_set_max_width (e->clue, e->painter, max_width - (html_engine_get_left_border (e) + html_engine_get_right_border (e)));
+	}
 }
 
 static void
@@ -307,7 +309,6 @@ html_iframe_real_calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed
 {
 	HTMLIFrame *iframe;
 	HTMLEngine *e;
-	gint width, height;
 	gint old_width, old_ascent, old_descent;
 	
 	old_width = o->width;
@@ -321,21 +322,15 @@ html_iframe_real_calc_size (HTMLObject *o, HTMLPainter *painter, GList **changed
 		return TRUE;
 
 	if ((iframe->width < 0) && (iframe->height < 0)) {
-		e->width = o->max_width;
-		html_engine_calc_size (e, changed_objs);
-
-		height = html_engine_get_doc_height (e);
-		width = html_engine_get_doc_width (e);
-
-		gtk_widget_set_size_request (iframe->scroll, width, height);
-		gtk_widget_queue_resize (iframe->scroll);
-		
+		if (e->clue) {
+			html_engine_calc_size (e, changed_objs);
+			e->width = html_engine_get_doc_width (e);
+			e->height = html_engine_get_doc_height (e);
+		}
 		html_iframe_set_scrolling (iframe, GTK_POLICY_NEVER);
 
-		e->width = width;
-		e->height = height;
-		o->width = width;
-		o->ascent = height;
+		o->width = e->width;
+		o->ascent = e->height;
 		o->descent = 0;
 	} else
 		return (* HTML_OBJECT_CLASS (parent_class)->calc_size) (o, painter, changed_objs);
@@ -471,7 +466,7 @@ save (HTMLObject *s,
 		if (state->error || 
 		    !html_engine_save_output_buffer (state,
 						     html_engine_save_buffer_peek_text (buffer),
-						     html_engine_save_buffer_peek_text_len (buffer))) {
+						     html_engine_save_buffer_peek_text_bytes (buffer))) {
 			html_engine_save_buffer_free (buffer);
 			return FALSE;
 		}
@@ -524,7 +519,7 @@ save_plain (HTMLObject *s,
 		if (state->error || 
 		    !html_engine_save_output_buffer (state,
 						     html_engine_save_buffer_peek_text (buffer),
-						     html_engine_save_buffer_peek_text_len (buffer))) {
+						     html_engine_save_buffer_peek_text_bytes (buffer))) {
 			html_engine_save_buffer_free (buffer);
 			return FALSE;
 		}
@@ -599,16 +594,16 @@ html_iframe_init (HTMLIFrame *iframe,
 	gtk_html_set_default_content_type (new_html,
 					   parent_html->priv->content_type);
 	iframe->html = new_widget;
-	depth = gtk_html_set_iframe_parent (new_html, parent, HTML_OBJECT (iframe));
-	gtk_container_add (GTK_CONTAINER (scrolled_window), new_widget);
-	gtk_widget_show (new_widget);
-
 	iframe->url = g_strdup (src);
 	iframe->width = width;
 	iframe->height = height;
 	iframe->gdk_painter = NULL;
 	iframe->frameborder = border;
 	gtk_html_set_base (new_html, src);
+	depth = gtk_html_set_iframe_parent (new_html, parent, HTML_OBJECT (iframe));
+	gtk_container_add (GTK_CONTAINER (scrolled_window), new_widget);
+	gtk_widget_show (new_widget);
+
 	g_signal_connect (new_html, "url_requested", G_CALLBACK (iframe_url_requested), iframe);
 
 	if (depth < 10) {
