@@ -24,22 +24,27 @@
 #ifndef _HTMLTABLECELL_H_
 #define _HTMLTABLECELL_H_
 
+#include <gdk/gdkcolor.h>
+
 #include "htmlcluev.h"
 #include "htmlimage.h"
 
 #define HTML_TABLE_CELL(x) ((HTMLTableCell *)(x))
 #define HTML_TABLE_CELL_CLASS(x) ((HTMLTableCellClass *)(x))
-#define IS_HTML_TABLE_CELL(x) (HTML_OBJECT_TYPE (x) == HTML_TYPE_TABLECELL)
+#define HTML_IS_TABLE_CELL(x) (HTML_OBJECT_TYPE (x) == HTML_TYPE_TABLECELL)
 
 struct _HTMLTableCell {
 	HTMLClueV cluev;
 
 	gint fixed_width;
 	gint fixed_height;
+	gboolean percent_width  : 1;
+	gboolean percent_height : 1;
+	gboolean no_wrap;
+	gboolean heading;
 
 	gint rspan;
 	gint cspan;
-	gint padding;
 	gint row;
 	gint col;
 
@@ -55,34 +60,31 @@ struct _HTMLTableCellClass {
 	HTMLClueVClass cluev_class;
 };
 
-
 extern HTMLTableCellClass html_table_cell_class;
 
-
-void        html_table_cell_type_init        (void);
-void        html_table_cell_class_init       (HTMLTableCellClass *klass,
-					      HTMLType            type,
-					      guint               object_size);
-void        html_table_cell_init             (HTMLTableCell      *cell,
-					      HTMLTableCellClass *klass,
-					      gint                percent,
-					      gint                rs,
-					      gint                cs,
-					      gint                pad);
-HTMLObject *html_table_cell_new              (gint                percent,
-					      gint                rs,
-					      gint                cs,
-					      gint                pad);
-void        html_table_cell_set_fixed_width  (HTMLTableCell      *cell,
-					      gint                width);
-void        html_table_cell_set_fixed_height (HTMLTableCell      *cell,
-					      gint                height);
-void        html_table_cell_set_bg_pixmap    (HTMLTableCell      *cell,
-					      HTMLImagePointer   *imagePtr);
-void        html_table_cell_set_position     (HTMLTableCell      *cell,
-					      gint                row,
-					      gint                col);
-gint        html_table_cell_get_fixed_width  (HTMLTableCell      *cell,
-					      HTMLPainter        *painter);
-
+void        html_table_cell_type_init         (void);
+void        html_table_cell_class_init        (HTMLTableCellClass *klass,
+					       HTMLType            type,
+					       guint               object_size);
+void        html_table_cell_init              (HTMLTableCell      *cell,
+					       HTMLTableCellClass *klass,
+					       gint                rs,
+					       gint                cs,
+					       gint                pad);
+HTMLObject *html_table_cell_new               (gint                rs,
+					       gint                cs,
+					       gint                pad);
+void        html_table_cell_set_fixed_width   (HTMLTableCell      *cell,
+					       gint                width,
+					       gboolean            percented);
+void        html_table_cell_set_fixed_height  (HTMLTableCell      *cell,
+					       gint                height,
+					       gboolean            percented);
+void        html_table_cell_set_bg_pixmap     (HTMLTableCell      *cell,
+					       HTMLImagePointer   *imagePtr);
+void        html_table_cell_set_position      (HTMLTableCell      *cell,
+					       gint                row,
+					       gint                col);
+gint        html_table_cell_get_fixed_width   (HTMLTableCell      *cell,
+					       HTMLPainter        *painter);
 #endif /* _HTMLTABLECELL_H_ */
